@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import dataclasses
 import os
+import warnings
 from typing import TYPE_CHECKING, ClassVar
 
 from loguru import logger
@@ -17,15 +18,14 @@ class _RunnerConfig:
 @dataclasses.dataclass(frozen=True)
 class _PyRunnerConfig(_RunnerConfig):
     name = "py"
+    use_thread_pool: bool | None
 
 
 @dataclasses.dataclass(frozen=True)
 class _RayRunnerConfig(_RunnerConfig):
     name = "ray"
     address: str | None
-    max_tasks_per_core: float | None
-    max_refs_per_core: float | None
-    batch_dispatch_coeff: float | None
+    max_task_backlog: int | None
 
 
 def _get_runner_config_from_env() -> _RunnerConfig:
@@ -36,22 +36,18 @@ def _get_runner_config_from_env() -> _RunnerConfig:
     1. PyRunner: set DAFT_RUNNER=py
     2. RayRunner: set DAFT_RUNNER=ray and optionally DAFT_RAY_ADDRESS=ray://...
     """
-    if "DAFT_RUNNER" in os.environ:
-        runner = os.environ["DAFT_RUNNER"]
-        if runner.upper() == "RAY":
-            tasks_per_core_env = os.getenv("DAFT_DEVELOPER_RAY_MAX_TASKS_PER_CORE")
-            refs_per_core_env = os.getenv("DAFT_DEVELOPER_RAY_MAX_REFS_PER_CORE")
-            batch_dispatch_env = os.getenv("DAFT_DEVELOPER_RAY_BATCH_DISPATCH_COEFF")
-            return _RayRunnerConfig(
-                address=os.getenv("DAFT_RAY_ADDRESS"),
-                max_tasks_per_core=float(tasks_per_core_env) if tasks_per_core_env else None,
-                max_refs_per_core=float(refs_per_core_env) if refs_per_core_env else None,
-                batch_dispatch_coeff=float(batch_dispatch_env) if batch_dispatch_env else None,
-            )
-        elif runner.upper() == "PY":
-            return _PyRunnerConfig()
-        raise ValueError(f"Unsupported DAFT_RUNNER variable: {os.environ['DAFT_RUNNER']}")
-    return _PyRunnerConfig()
+    runner = os.getenv("DAFT_RUNNER") or "PY"
+    if runner.upper() == "RAY":
+        task_backlog_env = os.getenv("DAFT_DEVELOPER_RAY_MAX_TASK_BACKLOG")
+        return _RayRunnerConfig(
+            address=os.getenv("DAFT_RAY_ADDRESS"),
+            max_task_backlog=int(task_backlog_env) if task_backlog_env else None,
+        )
+    elif runner.upper() == "PY":
+        use_thread_pool_env = os.getenv("DAFT_DEVELOPER_USE_THREAD_POOL")
+        use_thread_pool = bool(int(use_thread_pool_env)) if use_thread_pool_env is not None else None
+        return _PyRunnerConfig(use_thread_pool=use_thread_pool)
+    raise ValueError(f"Unsupported DAFT_RUNNER variable: {runner}")
 
 
 # Global Runner singleton, initialized when accessed through the DaftContext
@@ -76,15 +72,14 @@ class DaftContext:
             assert isinstance(self.runner_config, _RayRunnerConfig)
             _RUNNER = RayRunner(
                 address=self.runner_config.address,
-                max_tasks_per_core=self.runner_config.max_tasks_per_core,
-                max_refs_per_core=self.runner_config.max_refs_per_core,
-                batch_dispatch_coeff=self.runner_config.batch_dispatch_coeff,
+                max_task_backlog=self.runner_config.max_task_backlog,
             )
         elif self.runner_config.name == "py":
             from daft.runners.pyrunner import PyRunner
 
             logger.info("Using PyRunner")
-            _RUNNER = PyRunner()
+            assert isinstance(self.runner_config, _PyRunnerConfig)
+            _RUNNER = PyRunner(use_thread_pool=self.runner_config.use_thread_pool)
 
         else:
             raise NotImplementedError(f"Runner config implemented: {self.runner_config.name}")
@@ -99,6 +94,10 @@ class DaftContext:
 
         return _RUNNER
 
+    @property
+    def is_ray_runner(self) -> bool:
+        return isinstance(self.runner_config, _RayRunnerConfig)
+
 
 _DaftContext = DaftContext()
 
@@ -109,9 +108,8 @@ def get_context() -> DaftContext:
 
 def set_runner_ray(
     address: str | None = None,
-    max_tasks_per_core: float | None = None,
-    max_refs_per_core: float | None = None,
-    batch_dispatch_coeff: float | None = None,
+    noop_if_initialized: bool = False,
+    max_task_backlog: int | None = None,
 ) -> DaftContext:
     """Set the runner for executing Daft dataframes to a Ray cluster
 
@@ -120,29 +118,37 @@ def set_runner_ray(
     1. DAFT_RUNNER=ray
     2. Optionally, DAFT_RAY_ADDRESS=ray://...
 
+    **This function will throw an error if called multiple times in the same process.**
+
     Args:
         address: Address to head node of the Ray cluster. Defaults to None.
+        noop_if_initialized: If set to True, only the first call to this function will have any effect in setting the Runner.
+            Subsequent calls will have no effect at all. Defaults to False, which throws an error if this function is called
+            more than once per process.
 
     Returns:
         DaftContext: Daft context after setting the Ray runner
     """
     global _DaftContext
     if _DaftContext.disallow_set_runner:
+        if noop_if_initialized:
+            warnings.warn(
+                "Calling daft.context.set_runner_ray(noop_if_initialized=True) multiple times has no effect beyond the first call."
+            )
+            return _DaftContext
         raise RuntimeError("Cannot set runner more than once")
     _DaftContext = dataclasses.replace(
         _DaftContext,
         runner_config=_RayRunnerConfig(
             address=address,
-            max_tasks_per_core=max_tasks_per_core,
-            max_refs_per_core=max_refs_per_core,
-            batch_dispatch_coeff=batch_dispatch_coeff,
+            max_task_backlog=max_task_backlog,
         ),
         disallow_set_runner=True,
     )
     return _DaftContext
 
 
-def set_runner_py() -> DaftContext:
+def set_runner_py(use_thread_pool: bool | None = None) -> DaftContext:
     """Set the runner for executing Daft dataframes to your local Python interpreter - this is the default behavior.
 
     Alternatively, users can set this behavior via an environment variable: DAFT_RUNNER=py
@@ -155,7 +161,7 @@ def set_runner_py() -> DaftContext:
         raise RuntimeError("Cannot set runner more than once")
     _DaftContext = dataclasses.replace(
         _DaftContext,
-        runner_config=_PyRunnerConfig(),
+        runner_config=_PyRunnerConfig(use_thread_pool=use_thread_pool),
         disallow_set_runner=True,
     )
     return _DaftContext
