@@ -89,25 +89,35 @@ def _get_physical_plan(node: LogicalPlan, psets: dict[str, list[PartitionT]]) ->
             return physical_plan.global_limit(child_plan, node)
 
         elif isinstance(node, logical_plan.Repartition):
-            # Translate PartitionScheme to the appropriate fanout instruction.
-            fanout_instruction: execution_step.FanoutInstruction
+            # Case: simple repartition (split)
+            if node._scheme == PartitionScheme.UNKNOWN:
+                return physical_plan.flatten_plan(
+                    physical_plan.split(
+                        child_plan,
+                        num_input_partitions=node._children()[0].num_partitions(),
+                        num_output_partitions=node.num_partitions(),
+                    )
+                )
+
+            # All other repartitions require shuffling.
+
+            # Do the fanout.
+            fanout_plan: physical_plan.InProgressPhysicalPlan
             if node._scheme == PartitionScheme.RANDOM:
-                fanout_instruction = execution_step.FanoutRandom(num_outputs=node.num_partitions())
+                fanout_plan = physical_plan.fanout_random(child_plan, node)
             elif node._scheme == PartitionScheme.HASH:
                 fanout_instruction = execution_step.FanoutHash(
-                    num_outputs=node.num_partitions(),
+                    _num_outputs=node.num_partitions(),
                     partition_by=node._partition_by,
                 )
+                fanout_plan = physical_plan.pipeline_instruction(
+                    child_plan,
+                    fanout_instruction,
+                    node.resource_request(),
+                )
             else:
                 raise RuntimeError(f"Unimplemented partitioning scheme {node._scheme}")
 
-            # Do the fanout.
-            fanout_plan = physical_plan.pipeline_instruction(
-                child_plan=child_plan,
-                pipeable_instruction=fanout_instruction,
-                resource_request=node.resource_request(),
-            )
-
             # Do the reduce.
             return physical_plan.reduce(
                 fanout_plan=fanout_plan,
@@ -135,6 +145,12 @@ def _get_physical_plan(node: LogicalPlan, psets: dict[str, list[PartitionT]]) ->
                 join=node,
             )
 
+        elif isinstance(node, logical_plan.Concat):
+            return physical_plan.concat(
+                top_plan=_get_physical_plan(left_child, psets),
+                bottom_plan=_get_physical_plan(right_child, psets),
+            )
+
         else:
             raise NotImplementedError(f"Unsupported plan type {node}")
 
