@@ -1,23 +1,26 @@
 from __future__ import annotations
 
 import itertools
+import pathlib
 from abc import abstractmethod
 from dataclasses import asdict, dataclass
 from enum import Enum, IntEnum
 from pprint import pformat
 from typing import Any, Generic, TypeVar
 
+import fsspec
+
 from daft.datasources import SourceInfo, StorageType
+from daft.datatype import DataType
 from daft.errors import ExpressionTypeError
-from daft.execution.operators import OperatorEnum
-from daft.expressions import CallExpression, Expression, ExpressionList, col
+from daft.expressions import Expression, ExpressionsProjection, col
+from daft.expressions.testing import expr_structurally_equal
 from daft.internal.treenode import TreeNode
-from daft.logical.field import Field
 from daft.logical.map_partition_ops import ExplodeOp, MapPartitionOp
 from daft.logical.schema import Schema
 from daft.resource_request import ResourceRequest
-from daft.runners.partitioning import PartitionCacheEntry, vPartition
-from daft.types import ExpressionType
+from daft.runners.partitioning import PartitionCacheEntry
+from daft.table import Table
 
 
 class OpLevel(IntEnum):
@@ -138,7 +141,6 @@ class LogicalPlan(TreeNode["LogicalPlan"]):
 
                     helper(child, depth=depth, index=index + 1, prefix=prefix, header=header)
             else:
-
                 connector = "└─"
                 middle_child_header = "─┬─"
 
@@ -167,7 +169,6 @@ class LogicalPlan(TreeNode["LogicalPlan"]):
         return "".join(builder)
 
     def _repr_helper(self, **fields: Any) -> str:
-
         fields_to_print: dict[str, Any] = {}
         if "output" not in fields:
             fields_to_print["output"] = self.schema()
@@ -176,14 +177,14 @@ class LogicalPlan(TreeNode["LogicalPlan"]):
         fields_to_print["partitioning"] = self.partition_spec()
         reduced_types = {}
         for k, v in fields_to_print.items():
-            if isinstance(v, ExpressionList):
-                v = v.exprs
+            if isinstance(v, ExpressionsProjection):
+                v = list(v)
             elif isinstance(v, Schema):
-                v = v.to_column_expressions().exprs
+                v = list([col(field.name) for field in v])
             elif isinstance(v, PartitionSpec):
                 v = asdict(v)
-                if isinstance(v["by"], ExpressionList):
-                    v["by"] = v["by"].exprs
+                if isinstance(v["by"], ExpressionsProjection):
+                    v["by"] = list(v["by"])
             reduced_types[k] = v
         to_render: list[str] = [f"{self.__class__.__name__}\n"]
         space = "    "
@@ -210,11 +211,13 @@ class TabularFilesScan(UnaryNode):
         *,
         schema: Schema,
         source_info: SourceInfo,
-        predicate: ExpressionList | None = None,
+        fs: fsspec.AbstractFileSystem | None,
+        predicate: ExpressionsProjection | None = None,
         columns: list[str] | None = None,
         filepaths_child: LogicalPlan,
         filepaths_column_name: str,
         num_partitions: int | None = None,
+        limit_rows: int | None = None,
     ) -> None:
         if num_partitions is None:
             num_partitions = filepaths_child.num_partitions()
@@ -224,16 +227,20 @@ class TabularFilesScan(UnaryNode):
         if predicate is not None:
             self._predicate = predicate
         else:
-            self._predicate = ExpressionList([])
+            self._predicate = ExpressionsProjection([])
 
         if columns is not None:
-            self._output_schema = Schema([schema[col] for col in columns])
+            self._output_schema = Schema._from_field_name_and_types(
+                [(schema[col].name, schema[col].dtype) for col in columns]
+            )
         else:
             self._output_schema = schema
 
         self._column_names = columns
         self._columns = self._schema
         self._source_info = source_info
+        self._fs = fs
+        self._limit_rows = limit_rows
 
         # TabularFilesScan has a single child node that provides the filepaths to read from.
         assert (
@@ -274,6 +281,7 @@ class TabularFilesScan(UnaryNode):
         return TabularFilesScan(
             schema=self.schema(),
             source_info=self._source_info,
+            fs=self._fs,
             predicate=self._predicate if self._predicate is not None else None,
             columns=self._column_names,
             filepaths_child=child,
@@ -285,6 +293,7 @@ class TabularFilesScan(UnaryNode):
         return TabularFilesScan(
             schema=self.schema(),
             source_info=self._source_info,
+            fs=self._fs,
             predicate=self._predicate,
             columns=self._column_names,
             filepaths_child=new_children[0],
@@ -296,7 +305,6 @@ class InMemoryScan(UnaryNode):
     def __init__(
         self, cache_entry: PartitionCacheEntry, schema: Schema, partition_spec: PartitionSpec | None = None
     ) -> None:
-
         if partition_spec is None:
             partition_spec = PartitionSpec(scheme=PartitionScheme.UNKNOWN, num_partitions=1)
 
@@ -336,9 +344,9 @@ class FileWrite(UnaryNode):
     def __init__(
         self,
         input: LogicalPlan,
-        root_dir: str,
+        root_dir: str | pathlib.Path,
         storage_type: StorageType,
-        partition_cols: ExpressionList | None = None,
+        partition_cols: ExpressionsProjection | None = None,
         compression: str | None = None,
     ) -> None:
         assert (
@@ -350,13 +358,9 @@ class FileWrite(UnaryNode):
         if partition_cols is not None:
             self._partition_cols = partition_cols
         else:
-            self._partition_cols = ExpressionList([])
-        for field in input.schema():
-            assert ExpressionType.is_primitive(
-                field.dtype
-            ), f"we can currently only write out primitive types, got: {field}"
+            self._partition_cols = ExpressionsProjection([])
 
-        schema = Schema([Field("file_path", ExpressionType.from_py_type(str))])
+        schema = Schema._from_field_name_and_types([("file_path", DataType.string())])
 
         super().__init__(schema, input.partition_spec(), op_level=OpLevel.PARTITION)
         self._register_child(input)
@@ -396,18 +400,18 @@ class FileWrite(UnaryNode):
 class Filter(UnaryNode):
     """Which rows to keep"""
 
-    def __init__(self, input: LogicalPlan, predicate: ExpressionList) -> None:
+    def __init__(self, input: LogicalPlan, predicate: ExpressionsProjection) -> None:
         super().__init__(input.schema(), partition_spec=input.partition_spec(), op_level=OpLevel.PARTITION)
         self._register_child(input)
 
         self._predicate = predicate
-        predicate_schema = predicate.to_schema(input.schema())
+        predicate_schema = predicate.resolve_schema(input.schema())
 
-        for i, resolved_field in enumerate(predicate_schema.fields.values()):
+        for resolved_field, predicate_expr in zip(predicate_schema, predicate):
             resolved_type = resolved_field.dtype
-            if resolved_type != ExpressionType.logical():
+            if resolved_type != DataType.bool():
                 raise ValueError(
-                    f"Expected expression {self._predicate.exprs[i]} to resolve to type LOGICAL, but received: {resolved_type}"
+                    f"Expected expression {predicate_expr} to resolve to type Boolean, but received: {resolved_type}"
                 )
 
     def __repr__(self) -> str:
@@ -436,10 +440,10 @@ class Projection(UnaryNode):
     def __init__(
         self,
         input: LogicalPlan,
-        projection: ExpressionList,
+        projection: ExpressionsProjection,
         custom_resource_request: ResourceRequest = ResourceRequest(),
     ) -> None:
-        schema = projection.to_schema(input.schema())
+        schema = projection.resolve_schema(input.schema())
         super().__init__(schema, partition_spec=input.partition_spec(), op_level=OpLevel.ROW)
         self._register_child(input)
         self._projection = projection
@@ -449,7 +453,7 @@ class Projection(UnaryNode):
         return self._custom_resource_request
 
     def __repr__(self) -> str:
-        return self._repr_helper(output=self._projection.exprs)
+        return self._repr_helper(output=list(self._projection))
 
     def required_columns(self) -> list[set[str]]:
         return [self._projection.required_columns()]
@@ -475,16 +479,18 @@ class Projection(UnaryNode):
 
 
 class Sort(UnaryNode):
-    def __init__(self, input: LogicalPlan, sort_by: ExpressionList, descending: list[bool] | bool = False) -> None:
+    def __init__(
+        self, input: LogicalPlan, sort_by: ExpressionsProjection, descending: list[bool] | bool = False
+    ) -> None:
         pspec = PartitionSpec(scheme=PartitionScheme.RANGE, num_partitions=input.num_partitions(), by=sort_by)
         super().__init__(input.schema(), partition_spec=pspec, op_level=OpLevel.GLOBAL)
         self._register_child(input)
         self._sort_by = sort_by
 
-        for e in self._sort_by:
-            dtype = e.resolve_type(input.schema())
-            if dtype in {ExpressionType.null(), ExpressionType.bytes(), ExpressionType.logical()}:
-                raise ExpressionTypeError(f"Cannot sort on expression {e} with type: {dtype}")
+        resolved_sort_by_schema = self._sort_by.resolve_schema(input.schema())
+        for f, sort_by_expr in zip(resolved_sort_by_schema, self._sort_by):
+            if f.dtype == DataType.null() or f.dtype == DataType.binary() or f.dtype == DataType.bool():
+                raise ExpressionTypeError(f"Cannot sort on expression {sort_by_expr} with type: {f.dtype}")
 
         if isinstance(descending, bool):
             self._descending = [descending for _ in self._sort_by]
@@ -504,8 +510,8 @@ class Sort(UnaryNode):
         return (
             isinstance(other, Sort)
             and self.schema() == other.schema()
-            and self._sort_by == self._sort_by
-            and self._descending == self._descending
+            and self._sort_by == other._sort_by
+            and self._descending == other._descending
         )
 
     def rebuild(self) -> LogicalPlan:
@@ -539,16 +545,12 @@ class MapPartition(UnaryNode, Generic[TMapPartitionOp]):
             and self._map_partition_op == other._map_partition_op
         )
 
-    def eval_partition(self, partition: vPartition) -> vPartition:
+    def eval_partition(self, partition: Table) -> Table:
         return self._map_partition_op.run(partition)
 
 
 class Explode(MapPartition[ExplodeOp]):
-    def __init__(self, input: LogicalPlan, explode_expressions: ExpressionList):
-        assert [
-            isinstance(e, CallExpression) and e._operator == OperatorEnum.EXPLODE for e in explode_expressions
-        ], "Expressions supplied to Explode LogicalPlan must be a CallExpression with OperatorEnum.EXPLODE"
-
+    def __init__(self, input: LogicalPlan, explode_expressions: ExpressionsProjection):
         map_partition_op = ExplodeOp(input.schema(), explode_columns=explode_expressions)
         super().__init__(
             input,
@@ -596,7 +598,7 @@ class LocalLimit(UnaryNode):
         return [{name: name for name in self.schema().column_names()}]
 
     def _local_eq(self, other: Any) -> bool:
-        return isinstance(other, LocalLimit) and self.schema() == other.schema() and self._num == self._num
+        return isinstance(other, LocalLimit) and self.schema() == other.schema() and self._num == other._num
 
     def rebuild(self) -> LogicalPlan:
         return LocalLimit(input=self._children()[0].rebuild(), num=self._num)
@@ -622,7 +624,7 @@ class GlobalLimit(UnaryNode):
         return [{name: name for name in self.schema().column_names()}]
 
     def _local_eq(self, other: Any) -> bool:
-        return isinstance(other, GlobalLimit) and self.schema() == other.schema() and self._num == self._num
+        return isinstance(other, GlobalLimit) and self.schema() == other.schema() and self._num == other._num
 
     def rebuild(self) -> LogicalPlan:
         return GlobalLimit(input=self._children()[0].rebuild(), num=self._num)
@@ -630,7 +632,7 @@ class GlobalLimit(UnaryNode):
 
 class LocalCount(UnaryNode):
     def __init__(self, input: LogicalPlan) -> None:
-        schema = Schema([Field("count", ExpressionType.integer())])
+        schema = Schema._from_field_name_and_types([("count", DataType.int64())])
         super().__init__(schema, partition_spec=input.partition_spec(), op_level=OpLevel.PARTITION)
         self._register_child(input)
 
@@ -670,12 +672,12 @@ class PartitionScheme(Enum):
 class PartitionSpec:
     scheme: PartitionScheme
     num_partitions: int
-    by: ExpressionList | None = None
+    by: ExpressionsProjection | None = None
 
 
 class Repartition(UnaryNode):
     def __init__(
-        self, input: LogicalPlan, partition_by: ExpressionList, num_partitions: int, scheme: PartitionScheme
+        self, input: LogicalPlan, partition_by: ExpressionsProjection, num_partitions: int, scheme: PartitionScheme
     ) -> None:
         pspec = PartitionSpec(
             scheme=scheme,
@@ -686,8 +688,8 @@ class Repartition(UnaryNode):
         self._register_child(input)
         self._partition_by = partition_by
         self._scheme = scheme
-        if scheme == PartitionScheme.RANDOM and len(partition_by.names) > 0:
-            raise ValueError("Can not pass in random partitioning and partition_by args")
+        if scheme in (PartitionScheme.RANDOM, PartitionScheme.UNKNOWN) and len(partition_by.to_name_set()) > 0:
+            raise ValueError(f"Can not pass in {scheme} and partition_by args")
 
     def __repr__(self) -> str:
         return self._repr_helper(
@@ -773,34 +775,34 @@ class LocalAggregate(UnaryNode):
     def __init__(
         self,
         input: LogicalPlan,
-        agg: list[tuple[Expression, str]],
-        group_by: ExpressionList | None = None,
+        agg: list[Expression],
+        group_by: ExpressionsProjection | None = None,
     ) -> None:
-        cols_to_agg = ExpressionList([e for e, _ in agg])
+        self._cols_to_agg = ExpressionsProjection(agg)
         self._group_by = group_by
-        required_cols = set(cols_to_agg.required_columns())
 
         if group_by is not None:
-            group_and_agg_cols = ExpressionList(group_by.exprs + [e for e, _ in agg])
-            schema = group_and_agg_cols.to_schema(input.schema())
-            required_cols = required_cols | set(group_by.required_columns())
+            group_and_agg_cols = ExpressionsProjection(list(group_by) + agg)
+            schema = group_and_agg_cols.resolve_schema(input.schema())
         else:
-            schema = cols_to_agg.to_schema(input.schema())
+            schema = self._cols_to_agg.resolve_schema(input.schema())
 
-        self._required_cols = required_cols
         super().__init__(schema, partition_spec=input.partition_spec(), op_level=OpLevel.PARTITION)
         self._register_child(input)
-        self._agg = [(e, op) for e, (_, op) in zip(cols_to_agg, agg)]
+        self._agg = agg
 
     def __repr__(self) -> str:
-        return self._repr_helper(agg=[e for e, _ in self._agg], group_by=self._group_by)
+        return self._repr_helper(agg=self._agg, group_by=self._group_by)
 
     def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
         assert len(new_children) == 1
         return LocalAggregate(new_children[0], agg=self._agg, group_by=self._group_by)
 
     def required_columns(self) -> list[set[str]]:
-        return [self._required_cols]
+        required_cols = set(self._cols_to_agg.required_columns())
+        if self._group_by is not None:
+            required_cols = required_cols | set(self._group_by.required_columns())
+        return [required_cols]
 
     def input_mapping(self) -> list[dict[str, str]]:
         if self._group_by is not None:
@@ -812,14 +814,14 @@ class LocalAggregate(UnaryNode):
         return (
             isinstance(other, LocalAggregate)
             and self.schema() == other.schema()
-            and all(l[0].is_eq(r[0]) and l[1] == r[1] for l, r in zip(self._agg, other._agg))
+            and all(expr_structurally_equal(l, r) for l, r in zip(self._agg, other._agg))
             and self._group_by == other._group_by
         )
 
     def rebuild(self) -> LogicalPlan:
         return LocalAggregate(
             input=self._children()[0].rebuild(),
-            agg=[(e, op) for e, op in self._agg],
+            agg=self._agg,
             group_by=self._group_by if self._group_by is not None else None,
         )
 
@@ -828,11 +830,10 @@ class LocalDistinct(UnaryNode):
     def __init__(
         self,
         input: LogicalPlan,
-        group_by: ExpressionList,
+        group_by: ExpressionsProjection,
     ) -> None:
-
         self._group_by = group_by
-        schema = group_by.to_schema(input.schema())
+        schema = group_by.resolve_schema(input.schema())
         super().__init__(schema, partition_spec=input.partition_spec(), op_level=OpLevel.PARTITION)
         self._register_child(input)
 
@@ -934,8 +935,8 @@ class Join(BinaryNode):
         self,
         left: LogicalPlan,
         right: LogicalPlan,
-        left_on: ExpressionList,
-        right_on: ExpressionList,
+        left_on: ExpressionsProjection,
+        right_on: ExpressionsProjection,
         how: JoinType = JoinType.INNER,
     ) -> None:
         assert len(left_on) == len(right_on), "left_on and right_on must match size"
@@ -947,13 +948,11 @@ class Join(BinaryNode):
         self._left_on = left_on
         self._right_on = right_on
 
-        for e in self._left_on:
-            if e.resolve_type(left.schema()) == ExpressionType.null():
-                raise ExpressionTypeError(f"Cannot join on null type expression: {e}")
-
-        for e in self._right_on:
-            if e.resolve_type(right.schema()) == ExpressionType.null():
-                raise ExpressionTypeError(f"Cannot join on null type expression: {e}")
+        for schema, exprs in ((left.schema(), self._left_on), (right.schema(), self._right_on)):
+            resolved_schema = exprs.resolve_schema(schema)
+            for f, expr in zip(resolved_schema, exprs):
+                if f.dtype == DataType.null():
+                    raise ExpressionTypeError(f"Cannot join on null type expression: {expr}")
 
         self._how = how
         output_schema: Schema
@@ -966,14 +965,14 @@ class Join(BinaryNode):
         elif how == JoinType.INNER:
             num_partitions = max(left.num_partitions(), right.num_partitions())
             right_drop_set = {r.name() for l, r in zip(left_on, right_on) if l.name() == r.name()}
-            left_columns = left.schema().to_column_expressions()
-            right_columns = ExpressionList([col(f.name) for f in right.schema() if f.name not in right_drop_set])
+            left_columns = ExpressionsProjection.from_schema(left.schema())
+            right_columns = ExpressionsProjection([col(f.name) for f in right.schema() if f.name not in right_drop_set])
             unioned_expressions = left_columns.union(right_columns, rename_dup="right.")
             self._left_columns = left_columns
-            self._right_columns = ExpressionList(unioned_expressions.exprs[len(self._left_columns.exprs) :])
+            self._right_columns = ExpressionsProjection(list(unioned_expressions)[len(self._left_columns) :])
             self._output_projection = unioned_expressions
-            output_schema = self._left_columns.to_schema(left.schema()).union(
-                self._right_columns.to_schema(right.schema())
+            output_schema = self._left_columns.resolve_schema(left.schema()).union(
+                self._right_columns.resolve_schema(right.schema())
             )
 
         left_pspec = PartitionSpec(scheme=PartitionScheme.HASH, num_partitions=num_partitions, by=self._left_on)
@@ -1030,3 +1029,45 @@ class Join(BinaryNode):
             right_on=self._right_on,
             how=self._how,
         )
+
+
+class Concat(BinaryNode):
+    def __init__(self, top: LogicalPlan, bottom: LogicalPlan):
+        assert top.schema() == bottom.schema()
+        self._top = top
+        self._bottom = bottom
+
+        new_partition_spec = PartitionSpec(
+            PartitionScheme.UNKNOWN,
+            num_partitions=(top.partition_spec().num_partitions + bottom.partition_spec().num_partitions),
+            by=None,
+        )
+
+        super().__init__(top.schema(), partition_spec=new_partition_spec, op_level=OpLevel.GLOBAL)
+        self._register_child(self._top)
+        self._register_child(self._bottom)
+
+    def __repr__(self) -> str:
+        return self._repr_helper(num_partitions=self.num_partitions())
+
+    def copy_with_new_children(self, new_children: list[LogicalPlan]) -> LogicalPlan:
+        assert len(new_children) == 2
+        return Concat(new_children[0], new_children[1])
+
+    def required_columns(self) -> list[set[str]]:
+        return [set(), set()]
+
+    def input_mapping(self) -> list[dict[str, str]]:
+        return [
+            {name: name for name in self._top.schema().column_names()},
+            {name: name for name in self._bottom.schema().column_names()},
+        ]
+
+    def _local_eq(self, other: Any) -> bool:
+        return isinstance(other, Concat) and self.schema() == other.schema()
+
+    def rebuild(self) -> LogicalPlan:
+        return Concat(
+            top=self._children()[0].rebuild(),
+            bottom=self._children()[1].rebuild(),
+        )
