@@ -1,14 +1,13 @@
 from __future__ import annotations
 
-import copy
-
 from loguru import logger
 
 from daft import resource_request
-from daft.expressions import col
+from daft.expressions import ExpressionsProjection, col
 from daft.internal.rule import Rule
 from daft.logical.logical_plan import (
     Coalesce,
+    Concat,
     Filter,
     GlobalLimit,
     Join,
@@ -22,7 +21,6 @@ from daft.logical.logical_plan import (
     TabularFilesScan,
     UnaryNode,
 )
-from daft.logical.schema import ExpressionList
 
 
 class PushDownPredicates(Rule[LogicalPlan]):
@@ -36,6 +34,7 @@ class PushDownPredicates(Rule[LogicalPlan]):
         for op in self._supported_unary_nodes:
             self.register_fn(Filter, op, self._filter_through_unary_node)
         self.register_fn(Filter, Join, self._filter_through_join)
+        self.register_fn(Filter, Concat, self._filter_through_concat)
 
     def _filter_through_projection(self, parent: Filter, child: Projection) -> LogicalPlan | None:
         """Pushes Filter through Projections, only if filter does not rely on any projected columns
@@ -49,11 +48,10 @@ class PushDownPredicates(Rule[LogicalPlan]):
         can_push_down = []
         can_not_push_down = []
         for pred in filter_predicate:
-            required_names = {e for e in pred.required_columns()}
+            required_names = {e for e in pred._required_columns()}
             if all(name in child_input_mapping for name in required_names):
-                pred = copy.deepcopy(pred)
                 for name in required_names:
-                    pred = pred._replace_column_with_expression(col(name), col(child_input_mapping[name]))
+                    pred = pred._replace_column_with_expression(name, col(child_input_mapping[name]))
                 can_push_down.append(pred)
             else:
                 can_not_push_down.append(pred)
@@ -61,7 +59,7 @@ class PushDownPredicates(Rule[LogicalPlan]):
             return None
         logger.debug(f"Pushing down Filter predicate {can_push_down} into {child}")
         pushed_down_filter = Projection(
-            input=Filter(grandchild, predicate=ExpressionList(can_push_down)),
+            input=Filter(grandchild, predicate=ExpressionsProjection(can_push_down)),
             projection=child._projection,
             custom_resource_request=child.resource_request(),
         )
@@ -69,7 +67,7 @@ class PushDownPredicates(Rule[LogicalPlan]):
         if len(can_not_push_down) == 0:
             return pushed_down_filter
         else:
-            return Filter(pushed_down_filter, ExpressionList(can_not_push_down))
+            return Filter(pushed_down_filter, ExpressionsProjection(can_not_push_down))
 
     def _filter_through_unary_node(self, parent: Filter, child: UnaryNode) -> LogicalPlan | None:
         """Pushes Filter through "supported" UnaryNodes (see: self._supported_unary_nodes)
@@ -81,6 +79,19 @@ class PushDownPredicates(Rule[LogicalPlan]):
         logger.debug(f"Pushing Filter {parent} through {child}")
         return child.copy_with_new_children([Filter(grandchild, parent._predicate)])
 
+    def _filter_through_concat(self, parent: Filter, child: Concat) -> LogicalPlan | None:
+        """Pushes a Filter through a Concat to its left/right children
+
+        Filter-Concat-Bottom-* -> Concat-Filter-Bottom-*
+        Filter-Concat-Top-* -> Concat-Filter-Top-*
+        """
+        top = child._children()[0]
+        bottom = child._children()[1]
+        return Concat(
+            top=Filter(top, parent._predicate),
+            bottom=Filter(bottom, parent._predicate),
+        )
+
     def _filter_through_join(self, parent: Filter, child: Join) -> LogicalPlan | None:
         """Pushes Filter through a Join to its left/right children
 
@@ -97,16 +108,14 @@ class PushDownPredicates(Rule[LogicalPlan]):
         left_push_down = []
         right_push_down = []
         for pred in filter_predicate:
-            required_names = pred.required_columns()
+            required_names = pred._required_columns()
             if all(name in left_input_mapping for name in required_names):
-                pred = copy.deepcopy(pred)
                 for name in required_names:
-                    pred = pred._replace_column_with_expression(col(name), col(left_input_mapping[name]))
+                    pred = pred._replace_column_with_expression(name, col(left_input_mapping[name]))
                 left_push_down.append(pred)
             elif all(name in right_input_mapping for name in required_names):
-                pred = copy.deepcopy(pred)
                 for name in required_names:
-                    pred = pred._replace_column_with_expression(col(name), col(right_input_mapping[name]))
+                    pred = pred._replace_column_with_expression(name, col(right_input_mapping[name]))
                 right_push_down.append(pred)
             else:
                 can_not_push_down.append(pred)
@@ -117,16 +126,16 @@ class PushDownPredicates(Rule[LogicalPlan]):
 
         if len(left_push_down) > 0:
             logger.debug(f"Pushing down Filter predicate left side: {left_push_down} into Join")
-            left = Filter(left, predicate=ExpressionList(left_push_down))
+            left = Filter(left, predicate=ExpressionsProjection(left_push_down))
         if len(right_push_down) > 0:
             logger.debug(f"Pushing down Filter predicate right side: {right_push_down} into Join")
-            right = Filter(right, predicate=ExpressionList(right_push_down))
+            right = Filter(right, predicate=ExpressionsProjection(right_push_down))
 
         new_join = child.copy_with_new_children([left, right])
         if len(can_not_push_down) == 0:
             return new_join
         else:
-            return Filter(new_join, ExpressionList(can_not_push_down))
+            return Filter(new_join, ExpressionsProjection(can_not_push_down))
 
     @property
     def _supported_unary_nodes(self) -> set[type[UnaryNode]]:
@@ -164,7 +173,7 @@ class PruneColumns(Rule[LogicalPlan]):
             [
                 Projection(
                     grandchild,
-                    projection=ExpressionList(new_child_exprs),
+                    projection=ExpressionsProjection(new_child_exprs),
                     custom_resource_request=child.resource_request(),
                 )
             ]
@@ -176,17 +185,17 @@ class PruneColumns(Rule[LogicalPlan]):
         Projection-LocalAggregate-* -> Projection-<LocalAggregate with pruned columns>-*
         """
         parent_required_set = parent.required_columns()[0]
-        agg_op_pairs = child._agg
-        agg_ids = {e.name() for e, _ in agg_op_pairs}
+        agg_exprs = child._agg
+        agg_ids = {e.name() for e in agg_exprs}
         if agg_ids.issubset(parent_required_set):
             return None
 
-        new_agg_pairs = [(e, op) for e, op in agg_op_pairs if e.name() in parent_required_set]
+        new_agg_exprs = [e for e in agg_exprs if e.name() in parent_required_set]
         grandchild = child._children()[0]
 
         logger.debug(f"Pruning Columns:  {agg_ids - parent_required_set} in projection aggregate")
         return parent.copy_with_new_children(
-            [LocalAggregate(input=grandchild, agg=new_agg_pairs, group_by=child._group_by)]
+            [LocalAggregate(input=grandchild, agg=new_agg_exprs, group_by=child._group_by)]
         )
 
     def _aggregate_logical_plan(self, parent: LocalAggregate, child: LogicalPlan) -> LogicalPlan | None:
@@ -247,7 +256,7 @@ class PruneColumns(Rule[LogicalPlan]):
 
         return Projection(
             child,
-            projection=ExpressionList([e.to_column_expression() for e in child.schema() if e.name in parent_name_set]),
+            projection=ExpressionsProjection([col(f.name) for f in child.schema() if f.name in parent_name_set]),
         )
 
 
@@ -308,6 +317,29 @@ class PushDownClausesIntoScan(Rule[LogicalPlan]):
         super().__init__()
         # self.register_fn(Filter, TabularFilesScan, self._push_down_predicates_into_scan)
         self.register_fn(Projection, TabularFilesScan, self._push_down_projections_into_scan)
+        self.register_fn(LocalLimit, TabularFilesScan, self._push_down_local_limit_into_scan)
+
+    def _push_down_local_limit_into_scan(self, parent: LocalLimit, child: TabularFilesScan) -> LogicalPlan | None:
+        """Pushes LocalLimit into the limit_rows option of a TabularFilesScan.
+
+        LocalLimit(n)-TabularFilesScan-* -> TabularFilesScan(limit_rows=n)-*
+        """
+        if child._limit_rows is not None:
+            new_limit_rows = min(child._limit_rows, parent._num)
+        else:
+            new_limit_rows = parent._num
+
+        new_scan = TabularFilesScan(
+            schema=child._schema,
+            predicate=child._predicate,
+            columns=child._column_names,
+            source_info=child._source_info,
+            fs=child._fs,
+            filepaths_child=child._filepaths_child,
+            filepaths_column_name=child._filepaths_column_name,
+            limit_rows=new_limit_rows,
+        )
+        return new_scan
 
     def _push_down_projections_into_scan(self, parent: Projection, child: TabularFilesScan) -> LogicalPlan | None:
         """Pushes Projections into a scan as selected columns. Retains the Projection if there are non-column expressions.
@@ -319,16 +351,18 @@ class PushDownClausesIntoScan(Rule[LogicalPlan]):
         scan_columns = child.schema()
         if required_columns == scan_columns.to_name_set():
             return None
+        ordered_required_columns = [f.name for f in scan_columns if f.name in required_columns]
 
         new_scan = TabularFilesScan(
             schema=child._schema,
             predicate=child._predicate,
-            columns=list(required_columns),
+            columns=ordered_required_columns,
             source_info=child._source_info,
+            fs=child._fs,
             filepaths_child=child._filepaths_child,
             filepaths_column_name=child._filepaths_column_name,
         )
-        if any(not e.is_column() for e in parent._projection):
+        if any(not e._is_column() for e in parent._projection):
             return parent.copy_with_new_children([new_scan])
         else:
             return new_scan
@@ -358,19 +392,18 @@ class FoldProjections(Rule[LogicalPlan]):
 
             new_exprs = []
             for e in parent_projection:
-                if e.is_column():
+                if e._is_column():
                     name = e.name()
                     assert name is not None
                     e = child_projection.get_expression_by_name(name)
                 else:
-                    e = copy.deepcopy(e)
-                    to_replace = e.required_columns()
+                    to_replace = e._required_columns()
                     for name in to_replace:
-                        e = e._replace_column_with_expression(col(name), child_projection.get_expression_by_name(name))
+                        e = e._replace_column_with_expression(name, child_projection.get_expression_by_name(name))
                 new_exprs.append(e)
             return Projection(
                 grandchild,
-                ExpressionList(new_exprs),
+                ExpressionsProjection(new_exprs),
                 custom_resource_request=resource_request.ResourceRequest.max_resources(
                     [parent.resource_request(), child.resource_request()]
                 ),
@@ -392,7 +425,7 @@ class DropProjections(Rule[LogicalPlan]):
         parent_projection = parent._projection
         child_output = child.schema()
         if (
-            all(expr.is_column() for expr in parent_projection)
+            all(expr._is_column() for expr in parent_projection)
             and len(parent_projection) == len(child_output)
             and all(p.name() == c.name for p, c in zip(parent_projection, child_output))
         ):
