@@ -4,7 +4,6 @@ from dataclasses import dataclass
 
 from daft.dataframe.preview import DataFramePreview
 from daft.logical.schema import Schema
-from daft.viz.repr import vpartition_repr, vpartition_repr_html
 
 HAS_PILLOW = False
 try:
@@ -22,6 +21,7 @@ class DataFrameDisplay:
 
     preview: DataFramePreview
     schema: Schema
+    # These formatting options are deprecated for now and not guaranteed to be supported.
     column_char_width: int = 20
     max_col_rows: int = 3
     num_rows: int = 10
@@ -36,21 +36,29 @@ class DataFrameDisplay:
         return f"(Showing first {min(self.num_rows, len(self.preview.preview_partition))} of {self.preview.dataframe_num_rows} rows)"
 
     def _repr_html_(self) -> str:
-        return vpartition_repr_html(
-            self.preview.preview_partition,
-            self.schema,
-            self.num_rows,
-            self._get_user_message(),
-            max_col_width=self.column_char_width,
-            max_lines=self.max_col_rows,
-        )
+        if len(self.schema) == 0:
+            return "<small>(No data to display: Dataframe has no columns)</small>"
+
+        res = "<div>\n"
+
+        if self.preview.preview_partition is not None:
+            res += self.preview.preview_partition._repr_html_()
+        else:
+            res += self.schema._repr_html_()
+
+        res += f"\n<small>{self._get_user_message()}</small>\n</div>"
+
+        return res
 
     def __repr__(self) -> str:
-        return vpartition_repr(
-            self.preview.preview_partition,
-            self.schema,
-            self.num_rows,
-            self._get_user_message(),
-            max_col_width=self.column_char_width,
-            max_lines=self.max_col_rows,
-        )
+        if len(self.schema) == 0:
+            return "(No data to display: Dataframe has no columns)"
+
+        if self.preview.preview_partition is not None:
+            res = repr(self.preview.preview_partition)
+        else:
+            res = repr(self.schema)
+
+        res += f"\n{self._get_user_message()}"
+
+        return res
