@@ -20,6 +20,13 @@ html_favicon = "_static/daft-favicon.png"
 # -- General configuration ---------------------------------------------------
 # https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
 
+# The name of a reST role (builtin or Sphinx extension) to use as the default role, that
+# is, for text marked up `like this`. This can be set to 'py:obj' to make `filter` a
+# cross-reference to the Python function “filter”. The default is None, which doesn’t
+# reassign the default role.
+
+default_role = "py:obj"
+
 extensions = [
     "sphinx_reredirects",
     "sphinx.ext.autodoc",
@@ -28,10 +35,10 @@ extensions = [
     "sphinx.ext.linkcode",
     "IPython.sphinxext.ipython_console_highlighting",
     "myst_nb",
+    "sphinx_copybutton",
 ]
 
 templates_path = ["_templates"]
-exclude_patterns = ["release_notes/_template.rst"]
 
 
 # -- Options for Notebook rendering
@@ -55,6 +62,16 @@ html_theme_options = {
     "logo_only": True,
 }
 
+# -- Copy button configuration
+
+# This pattern matches:
+# - Python Repl prompts (">>> ") and it's continuation ("... ")
+# - Bash prompts ("$ ")
+# - IPython prompts ("In []: ", "In [999]: ") and it's continuations
+#   ("  ...: ", "     : ")
+copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
+copybutton_prompt_is_regexp = True
+
 # -- Options for redirecting URLs
 redirects = {
     "learn/install": "../install.html",
@@ -62,6 +79,7 @@ redirects = {
     "learn/user_guides/types_and_ops": "intro-dataframes.html",
     "learn/user_guides/remote_cluster_execution": "scaling-up.html",
     "learn/quickstart": "learn/10-min.html",
+    "learn/10-min": "../10-min.html",
 }
 
 # Resolving code links to github
