@@ -2,16 +2,14 @@ from __future__ import annotations
 
 import pytest
 
-from daft import DataFrame
-from daft.expressions import ColumnExpression
-from daft.logical.field import Field
+import daft
+from daft.datatype import DataType
 from daft.logical.logical_plan import LogicalPlan
-from daft.types import ExpressionType
 
 
 @pytest.fixture(scope="function")
 def df():
-    return DataFrame.from_pydict({"foo": [1, 2, 3]})
+    return daft.from_pydict({"foo": [1, 2, 3]})
 
 
 def test_get_plan(df):
@@ -26,7 +24,11 @@ def test_num_partitions(df):
 
 
 def test_schema(df):
-    assert [f for f in df.schema()] == [Field("foo", ExpressionType.integer())]
+    fields = [f for f in df.schema()]
+    assert len(fields) == 1
+    [field] = fields
+    assert field.name == "foo"
+    assert field.dtype == DataType.int64()
 
 
 def test_column_names(df):
@@ -36,5 +38,4 @@ def test_column_names(df):
 def test_columns(df):
     assert len(df.columns) == 1
     [ex] = df.columns
-    assert isinstance(ex, ColumnExpression)
     assert ex.name() == "foo"
