@@ -1,10 +1,10 @@
 from __future__ import annotations
 
-from daft import DataFrame
+import daft
 
 
 def test_with_column(valid_data: list[dict[str, float]]) -> None:
-    df = DataFrame.from_pylist(valid_data)
+    df = daft.from_pylist(valid_data)
     expanded_df = df.with_column("bar", df["sepal_width"] + df["petal_length"])
     data = expanded_df.to_pydict()
     assert expanded_df.column_names == df.column_names + ["bar"]
@@ -12,7 +12,7 @@ def test_with_column(valid_data: list[dict[str, float]]) -> None:
 
 
 def test_stacked_with_columns(valid_data: list[dict[str, float]]):
-    df = DataFrame.from_pylist(valid_data)
+    df = daft.from_pylist(valid_data)
     df = df.select(df["sepal_length"])
     df = df.with_column("sepal_length_2", df["sepal_length"])
     df = df.with_column("sepal_length_3", df["sepal_length_2"])
